##############################
##2013 Prefresher
##Day 3 - Booleans and for loops
##Stephen Pettigrew
##Harvard University
##pettigrew@fas.harvard.edu
##############################


##### This R script introduces different types of booleans,
##### as well as if statement and for loops


###########################
## Loading our data
###########################

## Let's begin by reading in the pres_ga.dta file.
## Don't forget to set your working directory.

require(foreign)

setwd("c:/users/stephen/dropbox/gov prefresher/2014 prefresher/r tutorials/day 2 data")

pres <- read.dta("pres08_ga.dta")



## Also, read in the day3data.RData file.

load("day3data.RData")








################################################
#### Boolean Logic
################################################

## A boolean is a data type which takes one of two values, true or false.
## R has several operators which can be used to create booleans.

## The most basic is the double equals sign:

2 == 1+1

1/2 == .5

2/3 == .6666 # False because of imperfect rounding



## We can also use the double equals sign (or any other boolean operator)
## on vectors

vec <- c(1,2,3,4,5)

vec == 3



## Other common boolean operators are greater than and less than:

2 > 3

2 < 3

vec > 3


## There's also a greater/less than or equal to operators:

vec >= 3

vec <= 3



## Recall that R treats logical data (TRUE and FALSE) similar to
## numerical data

## Use the sum() function to get the number of elements in vec that are greater
## than or equal to 3. Do this in one line of code












## Answer
sum(vec >= 3)




#### Another key boolean operator is the exclamation point, which negates logical
#### statements.

2 == 3 #FALSE

2 != 3 #TRUE

2 > 3 #FALSE

!(2 > 3) #TRUE

!c(T, F, T,T,T, F)


## We can use logical statements with character data also:

pres$county == "APPLING"


## Use what you learned yesterday about using brackets to look at
## the rows of "pres" from Appling County












pres[pres$county == "APPLING",]

subset(pres, pres$county == "APPLING")


## We can even get fancier by extracting the dempct values for these counties.
## You can do this three different ways.
## 1:
# This way isn't as common among programmers. It doesn't look as pretty as 2
pres[pres$county == "APPLING",]$dempct

## 2:
# Note that there is no comma within the bracket. Why?
pres$dempct[pres$county == "APPLING"]

## 3:
pres[pres$county == "APPLING", "dempct"]






## AND and OR statements

## You can use the AND operator, &, to evaluate multiple logical statements

(2 == 2) & (3 == 2) #FALSE

(2 == 2) & (3 == 3) #TRUE


## Alternatively, you can use the OR operator, |
## | is on your keyboard underneath the backspace

(2 == 2) | (3 == 2) #TRUE

(2 == 8) | (3 == 2) #FALSE


pres[(pres$county == "APPLING" & pres$year == 2004),]


## Another very useful boolean-like operator is %in%.
## It checks whether the first argument is contained anywhere in the second argument

2 %in% c(1,2,3) ## True

4 %in% c(1,2,3) ## False


## Two other useful functions to use when generating booleans are any() and all().
## You can use them when a line of code generates several TRUEs and FALSEs.
## any() returns TRUE if there's a single TRUE value inside the parentheses.
## all() returns TRUE if every value inside the parentheses is TRUE

2 == c(1,2,3) ## FALSE TRUE FALSE

any(2 == c(1,2,3)) ## TRUE

all(2 == c(1,2,3)) ## FALSE




###########################################################
## Applying booleans using if/else statements
###########################################################

## Sometimes you want to perform an operation only if certain conditions are met.
## To accomplish this we use 'if' and 'else' statements

## An 'if' statement takes the form:
## if(logical boolean){function to perform if the boolean is TRUE}

x <- 100000000

if(x > 10000){
  "x is a really big number"
}

if(x < 10000){
  "x isn't that big"
} ## Nothing happens here because x < 10000 evaluates to false





## We can use else statements to deal with instances where the boolean
## evaluates to FALSE

## Note that using multiple if/else statements requires that you wrap the whole thing
## in curly brackets, {}:

x <- 1

if(x > 10000){
  "x is a really big number"
  }  else{
    "x isn't that big"
  }

## The ifelse() function can be used instead of 'if' and 'else' separately:

ifelse(x > 10000,
       "x is a really big number",
       "x isn't that big")


## Typically though, it's better to avoid ifelse(), since often you want to 
## evaluate multiple 'if' statements:

y <- "Math!"

if(y > 10000) {
  "y is a really big number"
} else if(y == 10000){
  "y equals 10000"
} else if(!is.numeric(y)){
  "y isn't even a number"
}  else{
  "y isn't that big"
}


## To code that same thing using ifelse(), you have to write some nasty code:
## Don't ever write something like this. It's way too hard to follow.

ifelse(!is.numeric(y),
       "y isn't even a number",
       ifelse(y == 10000,
              "y equals 10000",
              ifelse(y > 10000,
                     "y is a really big number",
                     "y isn't that big")))





###########################################################
## For loops
###########################################################


## A "for-loop" is a way of repeating a process a vast 
## number of times.  For each iteration of the loop
## R keeps an index number stored which can be handy.


#######################################
## Indexing in loops
#######################################

## To understand loops in any programming language, you first
## need to understand indexing in a loop

## Take this very simple loop:
## In this case, the loop is indexed by 'i'.
## The loop starts with i = 1, goes through all the lines of the loop
## then returns to the top of the loop and sets i = 2. The loop knows to
## stop when it reaches i = 10

for(i in 1:10){

  print(i)

}

## That loop's indexing is identical to this one:

for(i in seq(1,10)){
  
  print(i)

}


## And this one:

index <- 1:10

for(i in index){

  print(i)

}





## Your indexes don't need to be sequential
# This one prints all even numbers between 2 and 50

for(i in seq(2,50,2)){
  
  print(i)
}


## Side note: "print()" statements are an incredibly useful
## thing to put into code. They'll help you debug your code by letting
## you know where your code hit snags. Or if you're looping over a huge number
## of things that take a really long time, they'll tell you where to restart
## your loop if you turn off your computer






## Your loop doesn't even need to be indexed by numbers:
index <- c("look", "how", "fancy", "we", "can",
           "be", "with", "loop", "indexes")

for(i in index){
  
  print(i)

}





## Another note about indexing and loops:
## If you put loops within loops, be careful that you never use the 
## the same letter to index two different loops.

## In other words, don't do this:

for(i in 1:10){
  
  for(i in 50:60){
  
    ## Some sort of code...
  }
  
}

## Don't this won't necessarily prevent your code from working,
## but it could cause you a headache that can be frustrating to 
## try to diagnose. 

## Instead your loops should look something like this:

for(i in 1:10){
  
  for(j in 50:60){
  
    ## some sort of code...
  }
  
}











## Write a for loop that takes the indexing 1:50, and
## prints each number in that range that is divisible
## by three.

## Hint: you'll probably want to use an if() statement
## as well as the modulus function: %%

for(i in 1:50){

}


















for(i in 1:50){
  if((i %% 3) == 0){
    print(i)
  }
}


## Note about the modulus function: what it does is divides the number
## in front of it by the number after it, and gives you the remainder.

## It's really useful to use inside of loops because you can use it to
## print the iteration you're on.

## In other words, if you have a loop that has 100,000 iterations, 
## maybe you want to keep tabs on how long it's taking and where it is,
## but don't want to print every single number from 1 to 100,000.
## Inserting this code into the loop will print only every 100th iteration:

if(i %% 100 == 0){
  print(i)
}





## Now let's get a little fancier with our loops.
## Let's write a loop that calculates and stores the average of the democratic
## vote percentage in each Georgia county in 2004 and 2008.

## First let's get the list of all the county names.

## How many counties are there in Georgia? Hint: use the unique() function














county.names <- unique(pres$county)

length(county.names)

## Fun fact: that's the second most of any state, after Texas.
## It's also the maximum number of counties allowed by the Georgia Constitution.
## The more you know... http://sighroll.com/wp-content/uploads/2012/10/themoreyouknow.jpg



## Now let's use our list of counties to get the 2004 & 2008 average democratic vote in
## each of the counties.

## Since we want to store the results to look at after the loop is completed, we need
## to first define a placeholder for that data.
## You have to make a placeholder everytime you calculate something inside of a loop
## that you want to access after the loop is done.

## Let's make a dataframe to store the means in.

## First we'll store the county names in the first column of the dataframe
county.means <- as.data.frame(county.names)

# Let's label that first column in the dataframe
colnames(county.means) <- "county"

## Now we'll make a placeholder for the county means
county.means$mean <- NA

## Let's look at what we have:
head(county.means)



## Note that in this case I chose to use "county" as the indexing argument, rather than i.
## It's typical to just use a single letter, but there's no reason you can't give it any
## other name, as long as you don't have another object in your workspace by that name.

for(county in county.names){
  ## We know we want to start the loop like this, but what goes inside here?
}



## When writing loops (or functions, or anything else) I find it often is easier
## to narrow the scope of what you're trying to do.
## Rather than thinking about all 159 counties at once, let's start easy and write
## the code that will calculate the 2004/2008 mean for the first county

county.names[1] ## APPLING

pres$dempct[pres$county == "APPLING"] ## This gives us the 2004 and 2008 pct. in Appling

mean(pres$dempct[pres$county == "APPLING"]) ## And there's the mean of 2004 and 2008

## And here's how we would store it in the dataframe:
county.means$mean[county.means$county == "APPLING"] <- 
  mean(pres$dempct[pres$county == "APPLING"])




## Now how would we do this for every county?

## Well let's look at what our indexes look like

for(county in county.names){
  print(county)
}



## Notice that each index is a character string (i.e. text and not a number)

for(i in county.names){
  
  internal.placeholder <- mean(pres$dempct[pres$county == i])
  
  county.means$mean[county.means$county == i] <- internal.placeholder

}





#####################################################
## When should we use loops, and when should we do something different?
#####################################################

## Answer: typically it's better to avoid using loops, if you can think
## of a way to write your code some other way.

## A common way to do this is called "vectorizing" your code.


## Think of matrix algebra.

## If we wanted to multiply "example.matrix" by its transpose, we could do so using loops:

example.matrix <- matrix(rnorm(5000), ncol = 5)

dim(example.matrix)

## What will be the dimensions of example.matrix times its transpose?










## Answer: 1000 x 1000


## How could we do the matrix multiplication using a loop?

result <- matrix(NA, 
                 nrow = 1000,
                 ncol = 1000)

## Look how long this takes!

for(i in 1:nrow(example.matrix)){

  for(j in 1:nrow(example.matrix)){
  
    result[i,j] <- sum(example.matrix[i,] * t(example.matrix)[,j])
  
  }
}









## If we had remembered what we know about matrix algebra, we would have vectoried
## our code and just used the canned R function that does matrix algebra:

asdf <- example.matrix %*% t(example.matrix) ## Way faster!









## Another example of why you want to vectorize code:

## Suppose at each iteration of a for loop, you want R to generate a random
## number and then use it to do calculations (this is the essence of Markov
## Chain Monte Carlo/ MCMC methods).

## Let's look at how much time you add to your computation if you draw a 
## random number in each iteration of loop versus vectorizing your code outside the loop.

## For a loop that does 10,000 iterations:


start <- proc.time()[3] ## this captures and stores the time before the loop starts

holder <- c()

for(i in 1:10000){
  holder[i] <- rnorm(1, mean = 0, sd = 1) 
  # here we are drawing one observation from a normal distribution
  # with mean zero and standard deviation 1
}

end <- proc.time()[3]


## How long did that take?

loop10000 <- end - start

loop10000



## What if we had vectorized our code and drawn all 10,000 random numbers first?

start <- proc.time()[3]

draws <-  rnorm(10000, mean = 0, sd = 1)

end <- proc.time()[3]

vect10000 <- end - start

loop10000
vect10000

## The vectorized code was way, way faster!











## And the time scales up very quickly. If we wanted a vector of 100,000 draws from
## a normal distribution...

start <- proc.time()[3]

holder <- c()

for(i in 1:100000){
  holder[i] <- rnorm(1, mean = 0, sd = 1)
}

end <- proc.time()[3]


## How long did that take?

loop.100000 <- end - start

loop.100000



## What if we had vectorized our code and drawn all 100,000 random numbers first?

start <- proc.time()[3]

draws <-  rnorm(100000, mean = 0, sd = 1)

end <- proc.time()[3]

vect.100000 <- end - start

loop.100000
vect.100000




## Why does it take so much longer?
## Think of it this way...

## Each step in the for loop draws a single random number, pulls the "holder" vector
## up from the computers memory, adds that single random number, and saves the "holder"
## vector to memory again.

## The vectorized code draws all 100,000 at once, and stores them all in one step

## Using a loop would be fine if you're only dealing with vectors with lengths 
## in the hundreds or a few thousand, but when you're talking 100,000+ 
## length vectors, computer memory is a serious problem.

## Moral of the story: remember your matrix algebra, vectorize your code, 
## and don't put things into for-loops if you can avoid it